#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
    Command-line for using nunavut and jinja to generate code
    from dsdl definitions.
"""

import argparse
import logging
import os
import pathlib
import sys
import textwrap
import typing


def _should_generate_support(args: argparse.Namespace) -> bool:
    if args.generate_support == 'as-needed':
        return (args.omit_serialization_support is None or not args.omit_serialization_support)
    else:
        return bool(args.generate_support == 'always' or args.generate_support == 'only')


def _run(args: argparse.Namespace, extra_includes: typing.List[str]) -> int:  # noqa: C901
    '''
        Post command-line setup and parsing logic to execute nunavut
        library routines based on input.
    '''
    #
    # nunavut : load module
    #
    import pydsdl
    import nunavut
    import nunavut.jinja
    import nunavut.lang

    def _build_ext_program_postprocessor(program: str, args: argparse.Namespace) \
            -> nunavut.postprocessors.FilePostProcessor:
        subprocess_args = [program]
        if hasattr(args, 'pp_run_program_arg') and args.pp_run_program_arg is not None:
            for program_arg in args.pp_run_program_arg:
                subprocess_args.append(program_arg)
        return nunavut.postprocessors.ExternalProgramEditInPlace(subprocess_args)

    def _build_post_processor_list_from_args(args: argparse.Namespace) \
            -> typing.List[nunavut.postprocessors.PostProcessor]:
        '''
        Return a list of post processors setup based on the provided command-line arguments. This
        list may be empty but the function will not return None.
        '''
        post_processors = []  # type: typing.List[nunavut.postprocessors.PostProcessor]
        if args.pp_trim_trailing_whitespace:
            post_processors.append(nunavut.postprocessors.TrimTrailingWhitespace())
        if hasattr(args, 'pp_max_emptylines') and args.pp_max_emptylines is not None:
            post_processors.append(nunavut.postprocessors.LimitEmptyLines(args.pp_max_emptylines))
        if hasattr(args, 'pp_run_program') and args.pp_run_program is not None:
            post_processors.append(_build_ext_program_postprocessor(args.pp_run_program, args))

        post_processors.append(nunavut.postprocessors.SetFileMode(args.file_mode))

        return post_processors

    #
    # nunavut: language context.
    #
    language_options = dict()
    if args.target_endianness is not None:
        language_options['target_endianness'] = args.target_endianness
    language_options['omit_float_serialization_support'] = args.omit_float_serialization_support
    language_options['enable_serialization_asserts'] = args.enable_serialization_asserts
    language_options['enable_override_variable_array_capacity'] = args.enable_override_variable_array_capacity

    language_context = nunavut.lang.LanguageContext(
        args.target_language,
        args.output_extension,
        args.namespace_output_stem,
        omit_serialization_support_for_target=args.omit_serialization_support,
        language_options=language_options,
        include_experimental_languages=args.experimental_languages)

    #
    # nunavut: inferred target language from extension
    #
    if args.output_extension is not None and language_context.get_target_language() is None:

        inferred_target_language_name = None  # type: typing.Optional[str]
        for name, lang in language_context.get_supported_languages().items():
            extension = lang.get_config_value('extension', None)
            if extension is not None and extension == args.output_extension:
                inferred_target_language_name = name
                break

        if inferred_target_language_name is not None:
            logging.info('Inferring target language %s based on extension "%s".',
                         inferred_target_language_name, args.output_extension)
            language_context = nunavut.lang.LanguageContext(
                inferred_target_language_name,
                args.output_extension,
                args.namespace_output_stem,
                omit_serialization_support_for_target=args.omit_serialization_support,
                language_options=language_options)
        elif args.templates is None:
            logging.warn(
                textwrap.dedent('''
                ***********************************************************************
                    No target language was given, none could be inferred from the output extension (-e) argument "%s",
                    and no user templates were specified. You will fail to find templates if you have provided any
                    DSDL types to generate.
                ***********************************************************************
                ''').lstrip(),
                args.output_extension
            )

    #
    # nunavut : parse
    #
    if args.generate_support != 'only':
        type_map = pydsdl.read_namespace(args.root_namespace,
                                         extra_includes,
                                         allow_unregulated_fixed_port_id=args.allow_unregulated_fixed_port_id)
    else:
        type_map = []

    root_namespace = nunavut.build_namespace_tree(
        type_map,
        args.root_namespace,
        args.outdir,
        language_context)

    #
    # nunavut : generate
    #

    generator_args = {
        'generate_namespace_types': (nunavut.YesNoDefault.YES
                                     if args.generate_namespace_types
                                     else nunavut.YesNoDefault.DEFAULT),
        'templates_dir': (pathlib.Path(args.templates) if args.templates is not None else None),
        'trim_blocks': args.trim_blocks,
        'lstrip_blocks': args.lstrip_blocks,
        'post_processors': _build_post_processor_list_from_args(args)
    }

    from nunavut.generators import create_generators
    generator, support_generator = create_generators(root_namespace, **generator_args)

    if args.list_outputs:
        if args.generate_support != 'only':
            for output_path in generator.generate_all(is_dryrun=True):
                sys.stdout.write(str(output_path))
                sys.stdout.write(';')

        if _should_generate_support(args):
            for output_path in support_generator.generate_all(is_dryrun=True):
                sys.stdout.write(str(output_path))
                sys.stdout.write(';')
        return 0

    if args.list_inputs:
        if args.generate_support != 'only':
            for input_path in generator.get_templates():
                sys.stdout.write(str(input_path.resolve()))
                sys.stdout.write(';')
        if _should_generate_support(args):
            for input_path in support_generator.get_templates():
                sys.stdout.write(str(input_path.resolve()))
                sys.stdout.write(';')
        if args.generate_support != 'only':
            if generator.generate_namespace_types:
                for output_type, _ in root_namespace.get_all_types():
                    sys.stdout.write(str(output_type.source_file_path))
                    sys.stdout.write(';')
            else:
                for output_type, _ in root_namespace.get_all_datatypes():
                    sys.stdout.write(str(output_type.source_file_path))
                    sys.stdout.write(';')
        return 0

    if _should_generate_support(args):
        support_generator.generate_all(is_dryrun=args.dry_run,
                                       allow_overwrite=not args.no_overwrite)

    if args.generate_support != 'only':
        generator.generate_all(is_dryrun=args.dry_run,
                               allow_overwrite=not args.no_overwrite)

    # If docgen selected, warn about linked namespaces
    if args.target_language == 'html':
        if len(extra_includes) > 0:
            logging.warning(
                "Other lookup namespaces are linked in these generated docs. "
                "If you do not generate docs for these other namespaces as well, "
                "links to external data types could be broken (expansion will still work)."
            )

    return 0


class _LazyVersionAction(argparse._VersionAction):
    '''
    Changes argparse._VersionAction so we only load nunavut.version
    if the --version action is requested.
    '''

    def __call__(self,
                 parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace,
                 values: typing.Any,
                 option_string: typing.Optional[str] = None) -> None:
        from nunavut.version import __version__
        parser._print_message(__version__, sys.stdout)
        parser.exit()


class _NunavutArgumentParser(argparse.ArgumentParser):
    """
    Specialization of argparse.ArgumentParser to encapsulate inter-argument rules.
    """

    def parse_known_args(self,
                         args: typing.Optional[typing.Sequence[str]] = None,
                         namespace: typing.Optional[argparse.Namespace] = None) \
            -> typing.Tuple[argparse.Namespace, typing.List[str]]:
        parsed_args, argv = super().parse_known_args(args, namespace)
        self._post_process_args(parsed_args)
        return (parsed_args, argv)

    def _post_process_args(self, args: argparse.Namespace) -> None:
        """
        Applies rules between different arguments and handles other special cases.
        """
        if args.list_inputs is not None and args.target_language is None and args.output_extension is None:
            # This is a special case where we know we'll never actually use the output extension since
            # we are only listing the input files. All other cases require either an output extension or
            # a valid target language.
            setattr(args, 'output_extension', '.tmp')

        if args.omit_serialization_support and args.generate_support == 'always':
            self.error(textwrap.dedent('''
                Logic error: use of --omit-serialization-support and --generate-support=always

                You cannot both omit serialization support and require generation of support code.
            ''').lstrip())


def _make_parser() -> argparse.ArgumentParser:
    """
        Defines the command-line interface. Provided as a separate factory method to
        support sphinx-argparse documentation.
    """

    epilog = textwrap.dedent('''

        **Example Usage**::

            # This would include j2 templates for a folder named 'c_jinja'
            # and generate .h files into a directory named 'include' using
            # dsdl root namespaces found under a folder named 'dsdl'.

            nnvg --outdir include --templates c_jinja -e .h dsdl

    ''')

    parser = _NunavutArgumentParser(
        description='Generate code from UAVCAN DSDL using pydsdl and jinja2',
        epilog=epilog,
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('root_namespace',
                        default='.',
                        nargs='?',
                        help='A source directory with DSDL definitions.')

    parser.add_argument('--lookup-dir', '-I', default=[], action='append',
                        help=textwrap.dedent('''

        List of other namespace directories containing data type definitions that are
        referred to from the target root namespace. For example, if you are reading a
        vendor-specific namespace, the list of lookup directories should always include
        a path to the standard root namespace "uavcan", otherwise the types defined in
        the vendor-specific namespace won't be able to use data types from the standard
        namespace.

        Additional directories can also be specified through the environment variable
        UAVCAN_DSDL_INCLUDE_PATH, where the path entries are
        separated by colons ":"

    ''').lstrip())

    parser.add_argument('--verbose', '-v', action='count',
                        help='verbosity level (-v, -vv)')

    parser.add_argument('--version', action=_LazyVersionAction)

    parser.add_argument(
        '--outdir', '-O', default='nunavut_out', help='output directory')

    parser.add_argument('--templates',
                        help=textwrap.dedent('''

        Paths to a directory containing templates to use when generating code.

        Templates found under these paths will override the built-in templates for a
        given language. If no target language was provided and no template paths were
        provided then no source will be generated.

    ''').lstrip())

    def extension_type(raw_arg: str) -> str:
        if len(raw_arg) > 0 and not raw_arg.startswith('.'):
            return '.' + raw_arg
        else:
            return raw_arg

    parser.add_argument('--target-language', '-l',
                        help=textwrap.dedent('''

        Language support to install into the templates.

        If provided then the output extension (--e) can be inferred otherwise the output
        extension must be provided.

    ''').lstrip())

    parser.add_argument('--experimental-languages', '-Xlang',
                        action='store_true',
                        help=textwrap.dedent('''

        Activate languages with unstable, experimental support.

        By default, target languages where support is not finalized are not
        enabled when running nunavut, to make it clear that the code output
        may change in a non-backwards-compatible way in future versions, or
        that it might not even work yet.

    ''').lstrip())

    ext_required = '--list-inputs' not in sys.argv and '--target-language' not in sys.argv and '-l' not in sys.argv
    parser.add_argument('--output-extension', '-e', type=extension_type,
                        required=ext_required,
                        help='The extension to use for generated files.')

    parser.add_argument('--dry-run', '-d', action='store_true',
                        help='If True then no files will be generated.')

    parser.add_argument('--list-outputs', action='store_true',
                        help=textwrap.dedent('''
        Emit a semicolon-separated list of files.
        (implies --dry-run)
        Emits files that would be generated if invoked without --dry-run.
        This command is useful for integrating with CMake and other build
        systems that need a list of targets to determine if a rebuild is
        necessary.

    ''').lstrip())

    parser.add_argument('--generate-support', choices=['always', 'never', 'as-needed', 'only'],
                        default='as-needed',
                        help=textwrap.dedent('''
        Change the criteria used to enable or disable support code generation.

        as-needed (default) - generate support code if serialization is enabled.
        always - always generate support code.
        never - never generate support code.
        only - only generate support code.

    ''').lstrip())

    parser.add_argument('--list-inputs', action='store_true',
                        help=textwrap.dedent('''

        Emit a semicolon-separated list of files.
        (implies --dry-run)
        A list of files that are resolved given input arguments like templates.
        This command is useful for integrating with CMake and other build systems
        that need a list of inputs to determine if a rebuild is necessary.

    ''').lstrip())

    parser.add_argument('--generate-namespace-types',
                        action='store_true',
                        help=textwrap.dedent('''
        If enabled this script will generate source for namespaces.
        All namespaces including and under the root namespace will be treated as a
        pseudo-type and the appropriate template will be used. The generator will
        first look for a template with the stem "Namespace" and will then use the
        "Any" template if that is available. The name of the output file will be
        the default value for the --namespace-output-stem argument and can be
        changed using that argument.

    ''').lstrip())

    parser.add_argument('--omit-serialization-support',
                        '-pod',
                        action='store_true',
                        help=textwrap.dedent('''
        If provided then the types generated will be POD datatypes with no additional logic.
        By default types generated include serialization routines and additional support libraries,
        headers, or methods.

    ''').lstrip())

    parser.add_argument('--namespace-output-stem',
                        default=None,
                        help='The name of the file generated when --generate-namespace-types is provided.')

    parser.add_argument('--no-overwrite',
                        action='store_true',
                        help=textwrap.dedent('''

        By default, generated files will be silently overwritten by
        subsequent invocations of the generator. If this argument is specified an
        error will be raised instead preventing overwrites.

    ''').lstrip())

    parser.add_argument('--file-mode',
                        default=0o444,
                        type=lambda value: int(value, 0),
                        help=textwrap.dedent('''

        The file-mode each generated file is set to after it is created.
        Note that this value is interpreted using python auto base detection.
        Because of this, to provide an octal value, you'll need to prefix your
        literal with '0o' (e.g. --file-mode 0o664).

    ''').lstrip())

    parser.add_argument('--trim-blocks',
                        action='store_true',
                        help=textwrap.dedent('''

        If this is set to True the first newline after a block in a template
        is removed (block, not variable tag!).

    ''').lstrip())

    parser.add_argument('--lstrip-blocks',
                        action='store_true',
                        help=textwrap.dedent('''

        If this is set to True leading spaces and tabs are stripped from the
        start of a line to a block in templates.

    ''').lstrip())

    parser.add_argument('--allow-unregulated-fixed-port-id',
                        action='store_true',
                        help=textwrap.dedent('''

        Do not reject unregulated fixed port identifiers.
        This is a dangerous feature that must not be used unless you understand the
        risks. The background information is provided in the UAVCAN specification.

    ''').lstrip())

    parser.add_argument('--pp-max-emptylines',
                        type=int,
                        help=textwrap.dedent('''

        If provided this will suppress generation of additional consecutive
        empty lines beyond the limit set by this argument.

        Note that this will insert a line post-processor which may reduce
        performance. Consider using a code formatter on the generated output
        to enforce whitespace rules instead.

    ''').lstrip())

    parser.add_argument('--pp-trim-trailing-whitespace',
                        action='store_true',
                        help=textwrap.dedent('''

        Enables a line post-processor that will elide all whitespace at the
        end of each line.

        Note that this will insert a line post-processor which may reduce
        performance. Consider using a code formatter on the generated output
        to enforce whitespace rules instead.

    ''').lstrip())

    parser.add_argument('-pp-rp', '--pp-run-program',
                        help=textwrap.dedent('''

        Runs a program after each file is generated but before the file is
        set to read-only.

        example ::

            # invokes clang-format with the "in-place" argument on each file after it is
            # generated.

            nnvg --outdir include --templates c_jinja -e .h -pp-rp clang-format -pp-rpa=-i dsdl

    ''').lstrip())

    parser.add_argument('-pp-rpa', '--pp-run-program-arg',
                        action='append',
                        help=textwrap.dedent('''

        Additional arguments to provide to the program specified by --pp-run-program.
        The last argument will always be the path to the generated file.

    ''').lstrip())

    ln_opt_group = parser.add_argument_group('language options', description=textwrap.dedent('''

        Options passed through to templates as `language_options` on the target language.

        Note that these arguments are passed though without validation, have no effect on the Nunavut
        library, and may or may not be appropriate based on the target language and generator templates
        in use.
    ''').lstrip())

    ln_opt_group.add_argument('--target-endianness',
                              choices=['any', 'big', 'little'],
                              help=textwrap.dedent('''

        Specify the endianness of the target hardware. This allows serialization
        logic to be optimized for different CPU architectures.

    ''').lstrip())

    ln_opt_group.add_argument('--omit-float-serialization-support',
                              action='store_true',
                              help=textwrap.dedent('''

        Instruct support header generators to omit support for floating point operations
        in serialization routines. This will result in errors if floating point types are used,
        however; if you are working on a platform without IEEE754 support and do not use floating
        point types in your message definitions this option will avoid dead code or compiler
        errors in generated serialization logic.

    ''').lstrip())

    ln_opt_group.add_argument('--enable-serialization-asserts',
                              action='store_true',
                              help=textwrap.dedent('''

        Instruct support header generators to generate language-specific assert statements as part
        of serialization routines. By default the serialization logic generated may make assumptions
        based on documented requirements for calling logic that could expose a system to undefined
        behavior. The alternative, for languages that do not support exception handling, is to
        use assertions designed to halt a program rather than execute undefined logic.

    ''').lstrip())

    ln_opt_group.add_argument('--enable-override-variable-array-capacity',
                              action='store_true',
                              help=textwrap.dedent('''

        Instruct support header generators to add the possibility to override max capacity of a
        variable length array in serialization routines. This option will disable serialization
        buffer checks and add conditional compilation statements which violates MISRA.

    ''').lstrip())

    return parser


def main() -> int:
    """
        Main entry point for this program.
    """

    #
    # Parse the command-line arguments.
    #
    args = _make_parser().parse_args()

    #
    # Setup Python logging.
    #
    fmt = '%(message)s'
    level = {0: logging.WARNING, 1: logging.INFO,
             2: logging.DEBUG}.get(args.verbose or 0, logging.DEBUG)
    logging.basicConfig(stream=sys.stderr, level=level, format=fmt)

    logging.info('Running %s using sys.prefix: %s',
                 pathlib.Path(__file__).name, sys.prefix)

    #
    # Parse UAVCAN_DSDL_INCLUDE_PATH
    #
    extra_includes = args.lookup_dir

    try:
        extra_includes_from_env = os.environ['UAVCAN_DSDL_INCLUDE_PATH'].split(':')

        logging.info('Additional include directories from UAVCAN_DSDL_INCLUDE_PATH: %s',
                     str(extra_includes_from_env))
        extra_includes += extra_includes_from_env
    except KeyError:
        pass

    return _run(args, extra_includes)



#
# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Contains supporting C headers to distribute with generated types.
"""
import pathlib
import typing
import pkg_resources

__version__ = "1.0.0"
"""Version of the c support headers."""


def list_support_files() -> typing.Generator[pathlib.Path, None, None]:
    """
    Get a list of C support headers embedded in this package.

    .. invisible-code-block: python

        from nunavut.lang.c.support import list_support_files
        import pathlib
        support_file_count = 0

    .. code-block:: python

        for path in list_support_files():
            support_file_count += 1
            assert path.parent.stem == 'support'
            assert (path.suffix == '.h' or path.suffix == '.j2')

    .. invisible-code-block: python

        assert support_file_count > 0

    :return: A list of C support header resources.
    """
    for resource in pkg_resources.resource_listdir(__name__, '.'):
        if resource.endswith('.h') or resource.endswith('.j2'):
            yield pathlib.Path(pkg_resources.resource_filename(__name__, str(resource)))



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
    Filters for generating C++. All filters in this
    module will be available in the template's global namespace as ``cpp``.
"""

import functools
import io
import re
import typing

import pydsdl

from ...templates import (template_language_filter,
                          template_language_list_filter)
from .. import Language
from .._common import IncludeGenerator, TokenEncoder, UniqueNameGenerator
from ..c import _CFit
from ..c import filter_literal as c_filter_literal

DEFAULT_ARRAY_TYPE = 'std::array<{TYPE},{MAX_SIZE}>'


def _handle_stropping_or_encoding_failure(encoder: TokenEncoder,
                                          stropped: str,
                                          token_type: str,
                                          pending_error: RuntimeError) -> str:
    """
    If the generic stropping fails we take one last look to see if there is something c++-specific we can do.
    """
    # Note that this is imprecise because C++ does not allow identifiers to start with an underscore if they are in the
    # global namespace, however, we don't have an AST for the C++ we're generating so there's no way to know if the
    # given token is global. Since this library is about generating code from DSDL and no DSDL identifier should be in
    # the global namespace (except for the top-level namespace for the datatypes) we assume that the token is not.
    m = re.match(r'^_+([A-Z]?)', stropped)
    if m:
        # Resolve the conflict between C's global identifier rules and our desire to use
        # '_' as a stropping prefix:
        return '_{}{}'.format(m.group(1).lower(), stropped[m.end():])

    # we couldn't help after all. raise the pending error.
    raise pending_error


@functools.lru_cache(maxsize=None)
def get_token_encoder(language: Language) -> TokenEncoder:
    """
    Caching getter to ensure we don't have to recompile TokenEncoders for each filter invocation.
    """
    return TokenEncoder(language,
                        stropping_failure_handler=_handle_stropping_or_encoding_failure,
                        encoding_failure_handler=_handle_stropping_or_encoding_failure)


@template_language_filter(__name__)
def filter_constant_value(language: Language,
                          constant: pydsdl.Constant) -> str:
    """
    Renders the specified value of the specified type as a literal.
    """
    return c_filter_literal(language, constant.value.native_value, constant.data_type)


def filter_to_standard_bit_length(t: pydsdl.PrimitiveType) -> int:
    """
    Returns the nearest standard bit length of a type as an int.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_to_standard_bit_length
        import pydsdl

    .. code-block: python
        # Given
        I = pydsdl.UnsignedIntegerType(7, pydsdl.PrimitiveType.CastMode.TRUNCATED)

        # and
        template = '{{ I | to_standard_bit_length }}'

        # then
        rendered = '8'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_standard_bit_length, template, rendered, 'c', I=I)

    """
    return int(_CFit.get_best_fit(t.bit_length).value)


@template_language_filter(__name__)
def filter_id(language: Language,
              instance: typing.Any,
              id_type: str = 'any') -> str:
    """
    Filter that produces a valid C and/or C++ identifier for a given object. The encoding may not
    be reversible.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_id


    .. code-block:: python

        # Given
        I = 'I like c++'

        # and
        template = '{{ I | id }}'

        # then
        rendered = 'I_like_czX002BzX002B'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, 'cpp', I=I)


    .. code-block:: python

        # Given
        I = 'if'

        # and
        template = '{{ I | id }}'

        # then
        rendered = '_if'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, 'cpp', I=I)


    .. code-block:: python

        # Given
        I = 'I   really like \t coffee'

        # and
        template = '{{ I | id }}'

        # then
        rendered = 'I_really_like_coffee'

    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, 'cpp', I=I)

    :param any instance:        Any object or data that either has a name property or can be converted
                                to a string.
    :returns: A token that is a valid identifier for C and C++, is not a reserved keyword, and is transformed
              in a deterministic manner based on the provided instance.
    """
    if hasattr(instance, 'name'):
        raw_name = str(instance.name)  # type: str
    else:
        raw_name = str(instance)

    return get_token_encoder(language).strop(raw_name, id_type)


@template_language_filter(__name__)
def filter_type(language: Language,
                obj: typing.Any) -> str:
    """
    Tries to convert a Python object into a c++ typename.

    Will raise a ValueError if the object provided does not (yet) have an available conversion in this function.

    Currently supported types are string:

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_type

    .. code-block:: python

         # given
        template = '{{ "Any" | type }}'

        # then
        rendered = "const char* const"

    .. invisible-code-block: python

        jinja_filter_tester(filter_type, template, rendered, 'cpp')

    int:

    .. code-block:: python

         # given
        template = '{{ 123 | type }}'

        # then
        rendered = 'long long'

    .. invisible-code-block: python

        jinja_filter_tester(filter_type, template, rendered, 'cpp')

    and bool:

    .. code-block:: python

         # given
        template = '{{ True | type }}'

        # then
        rendered = 'bool'

    .. invisible-code-block: python

        jinja_filter_tester(filter_type, template, rendered, 'cpp')

    """

    if isinstance(obj, bool):
        return "bool"
    if isinstance(obj, int):
        return "long long"
    if isinstance(obj, str):
        return "const char* const"

    return filter_type_from_primitive(language, obj)


@template_language_filter(__name__)
def filter_open_namespace(language: Language,
                          full_namespace: str,
                          bracket_on_next_line: bool = True,
                          linesep: str = '\n') -> str:
    """
    Emits c++ opening namespace syntax parsed from a pydsdl "full_namespace",
    dot-separated  value.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_open_namespace
        T = lambda : None;
        setattr(T, 'full_namespace)', '')

    .. code-block:: python

        # Given
        T.full_namespace = 'uavcan.foo'

        # and
        template = '{{ T.full_namespace | open_namespace }}'

        # then
        rendered = '''namespace uavcan
        {
        namespace foo
        {'''

    .. invisible-code-block: python

        # stropping doesn't change our example here.

        lctx = configurable_language_context_factory({'nunavut.lang.cpp': {'enable_stropping': False}}, 'cpp')
        jinja_filter_tester(filter_open_namespace, template, rendered, lctx, T=T)

        lctx = configurable_language_context_factory({'nunavut.lang.cpp': {'enable_stropping': True}}, 'cpp')
        jinja_filter_tester(filter_open_namespace, template, rendered, lctx, T=T)

    :param str full_namespace: A dot-separated namespace string.
    :param bool bracket_on_next_line: If True (the default) then the opening
        brackets are placed on a newline after the namespace keyword.
    :param str linesep: The line-separator to use when emitting new lines.
                        By default this is ``\\n``.

    :returns: C++ namespace declarations with opening brackets.
    """

    with io.StringIO() as content:
        first = True
        for name in full_namespace.split('.'):
            if first:
                first = False
            else:
                content.write(linesep)
            content.write('namespace ')
            if language.enable_stropping:
                content.write(filter_id(language, name))
            else:
                content.write(name)
            if bracket_on_next_line:
                content.write(linesep)
            else:
                content.write(' ')
            content.write('{')
        return content.getvalue()


@template_language_filter(__name__)
def filter_close_namespace(language: Language,
                           full_namespace: str,
                           omit_comments: bool = False,
                           linesep: str = '\n') -> str:
    """
    Emits c++ closing namespace syntax parsed from a pydsdl "full_namespace",
    dot-separated  value.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_close_namespace
        T = lambda : None;
        setattr(T, 'full_namespace)', '')

    .. code-block:: python

        # Given
        T.full_namespace = 'uavcan.foo'

        # and
        template = '{{ T.full_namespace | close_namespace }}'

        # then
        rendered = '''} // namespace foo
        } // namespace uavcan'''

    .. invisible-code-block: python

        jinja_filter_tester(filter_close_namespace, template, rendered, 'cpp', T=T)


    :param str full_namespace: A dot-separated namespace string.
    :param bool omit_comments: If True then the comments following the closing
                               bracket are omitted.
    :param str linesep: The line-separator to use when emitting new lines.
                        By default this is ``\\n``

    :returns: C++ namespace declarations with opening brackets.
    """
    with io.StringIO() as content:
        first = True
        for name in reversed(full_namespace.split('.')):
            if first:
                first = False
            else:
                content.write(linesep)

            content.write('}')
            if not omit_comments:
                content.write(' // namespace ')
                if language.enable_stropping:
                    content.write(filter_id(language, name))
                else:
                    content.write(name)
        return content.getvalue()


@template_language_filter(__name__)
def filter_full_reference_name(language: Language, t: pydsdl.CompositeType) -> str:
    """
    Provides a string that is the full namespace, typename, major, and minor version for a given composite type.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_full_reference_name
        from unittest.mock import MagicMock
        import pydsdl

        my_obj = MagicMock()
        my_obj.has_parent_service = False
        my_obj.version = MagicMock()

    .. code-block:: python

        # Given a type with illegal characters for C++
        my_obj.full_name = 'any.int.2Foo'
        my_obj.version.major = 1
        my_obj.version.minor = 2

        # and
        template = '{{ my_obj | full_reference_name }}'

        # then, with stropping enabled
        rendered = 'any::_int::_2Foo_1_2'

    .. invisible-code-block: python

        my_obj.short_name = my_obj.full_name.split('.')[-1]
        my_obj.full_namespace = '.'.join(my_obj.full_name.split('.')[:-1])

        jinja_filter_tester(filter_full_reference_name, template, rendered, 'cpp', my_obj=my_obj)

    :param pydsdl.CompositeType t: The DSDL type to get the fully-resolved reference name for.
    """
    ns_parts = t.full_namespace.split('.')
    if language.enable_stropping:
        ns = list(map(functools.partial(filter_id, language), ns_parts))
    else:
        ns = ns_parts

    full_path = ns + [filter_short_reference_name(language, t)]
    return '::'.join(full_path)


@template_language_filter(__name__)
def filter_short_reference_name(language: Language, t: pydsdl.CompositeType) -> str:
    """
    Provides a string that is a shorted version of the full reference name. This type is unique only within its
    namespace.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_short_reference_name
        from unittest.mock import MagicMock
        import pydsdl

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.parent_service = None

    .. code-block:: python

        # Given a type with illegal C++ characters
        my_type.short_name = '2Foo'
        my_type.version.major = 1
        my_type.version.minor = 2

        # and
        template = '{{ my_type | short_reference_name }}'

        # then, with stropping enabled
        rendered = '_2Foo_1_2'

    .. invisible-code-block: python

        jinja_filter_tester(filter_short_reference_name, template, rendered, 'cpp', my_type=my_type)

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.parent_service = None

    :param pydsdl.CompositeType t: The DSDL type to get the reference name for.
    """
    short_name = '{short}_{major}_{minor}'.format(short=t.short_name, major=t.version.major, minor=t.version.minor)
    if language.enable_stropping:
        return filter_id(language, short_name)
    else:
        return short_name


@template_language_list_filter(__name__)
def filter_includes(language: Language,
                    t: pydsdl.CompositeType,
                    sort: bool = True) -> typing.List[str]:
    """
    Returns a list of all include paths for a given type.

    :param pydsdl.CompositeType t: The type to scan for dependencies.
    :param bool sort: If true the returned list will be sorted.
    :return: a list of include headers needed for a given type.
    """

    include_gen = IncludeGenerator(language,
                                   t,
                                   filter_id,
                                   filter_short_reference_name)
    return include_gen.generate_include_filepart_list(language.extension, sort)


@template_language_filter(__name__)
def filter_declaration(language: Language,
                       instance: pydsdl.Any) -> str:
    """
    Emit a declaration statement for the given instance.
    """
    if isinstance(instance, pydsdl.PrimitiveType) or isinstance(instance, pydsdl.VoidType):
        return filter_type_from_primitive(language, instance)
    elif isinstance(instance, pydsdl.VariableLengthArrayType):
        variable_array_type = language.get_option('variable_array_type')

        if not isinstance(variable_array_type, str):
            raise RuntimeError('variable_array_type language option was missing or invalid.')
        return variable_array_type.format(
            TYPE=filter_declaration(language, instance.element_type),
            MAX_SIZE=instance.capacity)
    elif isinstance(instance, pydsdl.ArrayType):
        return DEFAULT_ARRAY_TYPE.format(
            TYPE=filter_declaration(language, instance.element_type),
            MAX_SIZE=instance.capacity)
    else:
        return filter_full_reference_name(language, instance)


@template_language_filter(__name__)
def filter_definition_begin(language: Language, instance: pydsdl.CompositeType) -> str:
    """
    Emit the start of a definition statement for a composite type.

     .. invisible-code-block: python

        from nunavut.lang.cpp import filter_definition_begin
        from unittest.mock import MagicMock
        import pytest
        import pydsdl

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.parent_service = None

        with pytest.raises(ValueError):
            jinja_filter_tester(filter_definition_begin,
                                '{{ my_type | definition_begin }}',
                                '',
                                'cpp',
                                my_type=MagicMock())

    .. code-block:: python

        # Given a pydsdl.CompositeType "my_type":
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        # and
        template = '{{ my_type | definition_begin }}'

        # then
        rendered = 'struct Foo_1_0'

    .. invisible-code-block: python

        jinja_filter_tester(filter_definition_begin, template, rendered, 'cpp', my_type=my_type)

        my_union_type = MagicMock(spec=pydsdl.UnionType)
        my_union_type.version = MagicMock()
        my_union_type.parent_service = None

    .. code-block:: python

        # Also, given a pydsdl.UnionType "my_union_type":
        my_union_type.short_name = 'Foo'
        my_union_type.version.major = 1
        my_union_type.version.minor = 0

        # and
        union_template = '{{ my_union_type | definition_begin }}'

        # then
        rendered = 'struct Foo_1_0'

    .. invisible-code-block: python

        jinja_filter_tester(filter_definition_begin, union_template, rendered, 'cpp', my_union_type=my_union_type)

        my_service_type = MagicMock(spec=pydsdl.ServiceType)
        my_service_type.version = MagicMock()
        my_service_type.parent_service = None

    .. code-block:: python

        # Finally, given a pydsdl.ServiceType "my_service_type":
        my_service_type.short_name = 'Foo'
        my_service_type.version.major = 1
        my_service_type.version.minor = 0

        # and
        template = '{{ my_service_type | definition_begin }}'

        # then
        rendered = 'namespace Foo_1_0'

    .. invisible-code-block: python

        jinja_filter_tester(filter_definition_begin, template, rendered, 'cpp', my_service_type=my_service_type)

    """
    short_name = filter_short_reference_name(language, instance)
    if isinstance(instance, pydsdl.DelimitedType) or isinstance(instance, pydsdl.StructureType) \
            or isinstance(instance, pydsdl.UnionType):
        return 'struct {}'.format(short_name)
    elif isinstance(instance, pydsdl.ServiceType):
        return 'namespace {}'.format(short_name)
    else:
        raise ValueError('{} types cannot be redefined.'.format(type(instance).__name__))


@template_language_filter(__name__)
def filter_definition_end(language: Language, instance: pydsdl.CompositeType) -> str:
    """
    Emit the end of a definition statement for a composite type.

     .. invisible-code-block: python

        from nunavut.lang.cpp import filter_definition_end
        from unittest.mock import MagicMock
        import pytest
        import pydsdl


        with pytest.raises(ValueError):
            jinja_filter_tester(filter_definition_end, '{{ my_type | definition_end }}', '', 'cpp', my_type=MagicMock())

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        jinja_filter_tester(filter_definition_end, '{{ my_type | definition_end }}', ';', 'cpp', my_type=my_type)

        my_type = MagicMock(spec=pydsdl.UnionType)
        my_type.version = MagicMock()
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        jinja_filter_tester(filter_definition_end, '{{ my_type | definition_end }}', ';', 'cpp', my_type=my_type)

        my_type = MagicMock(spec=pydsdl.ServiceType)
        my_type.version = MagicMock()
        my_type.parent_service = None
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        jinja_filter_tester(filter_definition_end,
                            '{{ my_type | definition_end }}',
                            ' // namespace Foo_1_0',
                            'cpp',
                            my_type=my_type)

    """
    if isinstance(instance, pydsdl.DelimitedType) or isinstance(instance, pydsdl.StructureType) \
            or isinstance(instance, pydsdl.UnionType):
        return ';'
    elif isinstance(instance, pydsdl.ServiceType):
        return ' // namespace {}'.format(filter_short_reference_name(language, instance))
    else:
        raise ValueError('{} types cannot be redefined.'.format(type(instance).__name__))


@template_language_filter(__name__)
def filter_type_from_primitive(language: Language,
                               value: pydsdl.PrimitiveType) -> str:
    """
    Filter to transform a pydsdl :class:`~pydsdl.PrimitiveType` into
    a valid C++ type.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_type_from_primitive
        import pydsdl


    .. code-block:: python

        # Given
        template = '{{ unsigned_int_32_type | type_from_primitive }}'

        # then
        rendered = 'std::uint32_t'

    .. invisible-code-block: python

        test_type = pydsdl.UnsignedIntegerType(32, pydsdl.PrimitiveType.CastMode.TRUNCATED)
        jinja_filter_tester(filter_type_from_primitive,
                            template,
                            rendered,
                            'cpp',
                            unsigned_int_32_type=test_type)

    Also note that this is sensitive to the ``use_standard_types`` configuration in the language properties:

    .. code-block:: python

        # rendered will be different if use_standard_types is False
        rendered = 'unsigned long'

    .. invisible-code-block: python

        lctx = configurable_language_context_factory({'nunavut.lang.cpp': {'use_standard_types': False}}, 'cpp')
        jinja_filter_tester(filter_type_from_primitive,
                            template,
                            rendered,
                            lctx,
                            unsigned_int_32_type=test_type)

    :param str value: The dsdl primitive to transform.

    :returns: A valid C++ type name.

    :raises TemplateRuntimeError: If the primitive cannot be represented as a standard C++ type.
    """
    return _CFit.get_best_fit(value.bit_length).to_c_type(value, language, 'std::')


def filter_to_namespace_qualifier(namespace_list: typing.List[str]) -> str:
    """
    Converts a list of namespace names into a qualifier string. For example:

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_to_namespace_qualifier
        import pydsdl


    .. code-block:: python

        my_namespace = ['foo', 'bar']
        template = '{{ my_namespace | to_namespace_qualifier }}myType()'
        expected = 'foo::bar::myType()'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_namespace_qualifier, template, expected, 'cpp', my_namespace=my_namespace)

    This filter gracefully handles empty namespace lists:

    .. code-block:: python

        my_namespace = []
        template = '{{ my_namespace | to_namespace_qualifier }}myType()'
        expected = 'myType()'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_namespace_qualifier, template, expected, 'cpp', my_namespace=my_namespace)

    """
    if namespace_list is None or len(namespace_list) == 0:
        return ''
    else:
        return '::'.join(namespace_list) + '::'


def filter_to_template_unique_name(base_token: str) -> str:
    """
    Filter that takes a base token and forms a name that is very
    likely to be unique within the template the filter is invoked. This
    name is also very likely to be a valid C++ identifier.

    .. IMPORTANT::

        The exact tokens generated may change between major or minor versions
        of this library. The only guarantee provided is that the tokens
        will be stable for the same version of this library given the same
        input.

        Also note that name uniqueness is only likely within a given template.
        Between templates there is no guarantee of uniqueness and,
        since this library does not lex generated source, there is no guarantee
        that the generated name does not conflict with a name generated by
        another means.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_to_template_unique_name
        from nunavut.lang._common import UniqueNameGenerator

    .. code-block:: python

        # Given
        template  = '{{ "foo" | to_template_unique_name }},{{ "Foo" | to_template_unique_name }},'
        template += '{{ "fOO" | to_template_unique_name }}'

        # then
        rendered = '_foo0_,_foo1_,_fOO0_'

    .. invisible-code-block: python

        UniqueNameGenerator.reset()
        jinja_filter_tester(filter_to_template_unique_name, template, rendered, 'cpp')

    .. code-block:: python

        # Given
        template = '{{ "i like coffee" | to_template_unique_name }}'

        # then
        rendered = '_i like coffee0_'

    .. invisible-code-block: python

        UniqueNameGenerator.reset()
        jinja_filter_tester(filter_to_template_unique_name, template, rendered, 'cpp')


    :param str base_token: A token to include in the base name.
    :returns: A name that is likely to be valid C++ identifier and is likely to
        be unique within the file generated by the current template.
    """
    if len(base_token) > 0:
        adj_base_token = base_token[0:1].lower() + base_token[1:]
    else:
        adj_base_token = base_token

    return UniqueNameGenerator.get_instance()('cpp', adj_base_token, '_', '_')


def filter_as_boolean_value(value: bool) -> str:
    """
    Filter a boolean expression to produce a valid C++ "true" or "false" token.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_as_boolean_value

    .. code-block:: python

        assert "true" == filter_as_boolean_value(True)
        assert "false" == filter_as_boolean_value(False)

    """
    return ('true' if value else 'false')


@template_language_filter(__name__)
def filter_indent(language: Language,
                  text: str,
                  depth: int = 1) -> str:
    """
    Emit indent characters as configured for the language.

    :param text: The text to indent.
    :param depth: The number of indents. For example, if depth is 2 and the indent for this language is
        4 spaces then the text will be indented by 8 spaces.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_indent

    .. code-block:: python

        # Given a string with an existing indent of 4 spaces...
        template  = '{{ "    int a = 1;" | indent }}'

        # then if cpp.indent == 4 we expect no change.
        rendered = '    int a = 1;'

    .. invisible-code-block: python

        assert_language_config_value('cpp', 'indent', '4', 'this test expected an indent of 4')
        jinja_filter_tester(filter_indent, template, rendered, 'cpp')

    .. code-block:: python

        # If the indent is only 3 spaces...
        template  = '{{ "   int a = 1;" | indent }}'

        # then if cpp.indent == 4 we expect 4 spaces (i.e. not 7 spaces)
        rendered = '    int a = 1;'

    .. invisible-code-block: python

        jinja_filter_tester(filter_indent, template, rendered, 'cpp')

    .. code-block:: python

        # We can also specify multiple indents...
        template  = '{{ "int a = 1;" | indent(2) }}'

        rendered = '        int a = 1;'

    .. invisible-code-block: python

        jinja_filter_tester(filter_indent, template, rendered, 'cpp')

    .. code-block:: python

        # ...or no indent
        template  = '{{ "    int a = 1;" | indent(0) }}'

        rendered = 'int a = 1;'

    .. invisible-code-block: python

        jinja_filter_tester(filter_indent, template, rendered, 'cpp')


    """
    configured_indent = int(language.get_config_value('indent'))
    lines = text.splitlines(keepends=True)
    result = ''
    for i in range(0, len(lines)):
        line = lines[i].lstrip()
        result += (' ' * (depth * configured_indent)) + line
    return result


def filter_minimum_required_capacity_bits(t: pydsdl.SerializableType) -> int:
    """
    Returns the minimum number of bits required to store the deserialized value of a
    pydsdl :class:`~pydsdl.SerializableType`. This capacity may be too small for some
    instances of the value (e.g. variable length arrays).

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_minimum_required_capacity_bits
        import pydsdl
        import pytest

    .. code-block:: python

        # Given
        template = '{{ unsigned_int_32_type | minimum_required_capacity_bits }}'

        # then
        rendered = '32'

    .. invisible-code-block: python

        uint32_truncated_type = pydsdl.UnsignedIntegerType(32, pydsdl.PrimitiveType.CastMode.TRUNCATED)
        jinja_filter_tester(filter_minimum_required_capacity_bits,
                            template,
                            rendered,
                            'cpp',
                            unsigned_int_32_type=uint32_truncated_type)

        # Cover the other cases here. We don't need to pain the user with these gory
        # details so we'll keep them in the hidden block.
        assert filter_minimum_required_capacity_bits(pydsdl.VoidType(1)) == 1
        assert filter_minimum_required_capacity_bits(pydsdl.FixedLengthArrayType(uint32_truncated_type, 2)) == 64
        assert filter_minimum_required_capacity_bits(pydsdl.VariableLengthArrayType(uint32_truncated_type, 2)) == 8

        field_one = pydsdl.Field(pydsdl.UnsignedIntegerType(32, pydsdl.PrimitiveType.CastMode.TRUNCATED), 'one')
        field_two = pydsdl.Field(pydsdl.UnsignedIntegerType(64, pydsdl.PrimitiveType.CastMode.TRUNCATED), 'two')

        test_struct = pydsdl.StructureType(
            name='uavcan.foo',
            version=pydsdl.Version(0, 1),
            attributes=[field_one, field_two],
            deprecated=False,
            fixed_port_id=None,
            source_file_path='',
            has_parent_service = False
        )

        test_union = pydsdl.UnionType(
            name='uavcan.foo',
            version=pydsdl.Version(0, 1),
            attributes=[field_one, field_two],
            deprecated=False,
            fixed_port_id=None,
            source_file_path='',
            has_parent_service = False
        )

        assert filter_minimum_required_capacity_bits(test_struct) == 96
        assert filter_minimum_required_capacity_bits(test_union) == 32 + 8

    :param pydsdl.SerializableType t: The dsdl type.

    :returns: The minimum, required bits needed to store some values of the given type.
    """
    return typing.cast(int, min(t.bit_length_set))



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""Code generator built on top of pydsdl.

Nunavut uses pydsdl to generate text files using templates. While these
text files are often source code this module could also be used to generate
documentation or data interchange formats like JSON or XML.

The input to the nunavut library is a list of templates and a list of
``pydsdl.pydsdl.CompositeType`` objects. The latter is typically obtained
by calling pydsdl::

    from pydsdl import read_namespace

    compound_types = read_namespace(root_namespace, include_paths)

Next a :class:`nunavut.lang.LanguageContext` is needed which is used to
configure all Nunavut objects for a specific target language ::

    from nunavut.lang import LanguageContext

    # Here we are going to generate C headers.
    langauge_context = LanguageContext('c')

:class:`nunavut.generators.AbstractGenerator` objects require
a :class:`nunavut.Namespace` tree which can be built from the
pydsdl type map using :meth:`nunavut.build_namespace_tree`::

    from nunavut import build_namespace_tree

    root_namespace = build_namespace_tree(compound_types,
                                          root_ns_folder,
                                          out_dir,
                                          language_context)

Putting this all together, the typical use of this library looks something like this::

    from pydsdl import read_namespace
    from nunavut import build_namespace_tree
    from nunavut.lang import LanguageContext
    from nunavut.jinja import DSDLCodeGenerator

    # parse the dsdl
    compound_types = read_namespace(root_namespace, include_paths)

    # select a target language
    langauge_context = LanguageContext('c')

    # build the namespace tree
    root_namespace = build_namespace_tree(compound_types,
                                          root_ns_folder,
                                          out_dir,
                                          language_context)

    # give the root namespace to the generator and...
    generator = DSDLCodeGenerator(root_namespace)

    # generate all the code!
    generator.generate_all()

"""

import collections
import enum
import pathlib
import sys
import typing

import pydsdl

from . import lang

if sys.version_info[:2] < (3, 5):   # pragma: no cover
    print('A newer version of Python is required', file=sys.stderr)
    sys.exit(1)


@enum.unique
class YesNoDefault(enum.Enum):
    """
    Trinary type for decisions that allow a default behavior to be requested that can
    be different based on other contexts. For example:

    .. invisible-code-block: python

        from datetime import datetime
        from nunavut import YesNoDefault

    .. code-block:: python

        def should_we_order_pizza(answer: YesNoDefault) -> bool:
            if answer == YesNoDefault.YES or (
               answer == YesNoDefault.DEFAULT and
               datetime.today().isoweekday() == 5):
                # if yes or if we are taking the default action which is to
                # order pizza on Friday, and today is Friday, then we order pizza
                return True
            else:
                return False

    .. invisible-code-block: python

        assert should_we_order_pizza(YesNoDefault.YES)
        assert not should_we_order_pizza(YesNoDefault.NO)

    """
    NO = 0
    YES = 1
    DEFAULT = 2

# +---------------------------------------------------------------------------+


class Namespace(pydsdl.Any):
    """
    K-ary tree (where K is the largest set of data types in a single dsdl namespace) where
    the nodes represent dsdl namespaces and the children are the datatypes and other nested
    namespaces (with datatypes always being leaf nodes). This structure extends :code:`pydsdl.Any`
    and is a :code:`pydsdl.pydsdl.CompositeType` via duck typing.

    :param str full_namespace:  The full, dot-separated name of the namepace. This is expected to be
                                a unique identifier.
    :param pathlib.Path root_namespace_dir: The directory representing the dsdl namespace and containing the
                                namespaces's datatypes and nested namespaces.
    :param pathlib.PurePath base_output_path: The base path under which all namespaces and datatypes should
                                be generated.
    :param lang.LanguageContext language_context: The generated software language context the namespace is within.
    """

    DefaultOutputStem = '_'

    def __init__(self,
                 full_namespace: str,
                 root_namespace_dir: pathlib.Path,
                 base_output_path: pathlib.PurePath,
                 language_context: lang.LanguageContext):
        self._parent = None  # type: typing.Optional[Namespace]
        self._id_filter = language_context.get_target_id_filter()
        self._namespace_components = []  # type: typing.List[str]
        self._namespace_components_stropped = []  # type: typing.List[str]
        for component in full_namespace.split('.'):
            self._namespace_components_stropped.append(self._id_filter(component, 'typedef'))
            self._namespace_components.append(component)
        self._full_namespace = '.'.join(self._namespace_components_stropped)
        self._output_folder = pathlib.Path(base_output_path / pathlib.PurePath(*self._namespace_components_stropped))
        output_stem = language_context.get_default_namespace_output_stem()
        if output_stem is None:
            output_stem = self.DefaultOutputStem
        output_path = self._output_folder / pathlib.PurePath(output_stem)
        self._support_output_folder = base_output_path
        self._output_path = output_path.with_suffix(language_context.get_output_extension())
        self._source_folder = pathlib.Path(
            root_namespace_dir / pathlib.PurePath(*self._namespace_components[1:])).resolve()
        if not self._source_folder.exists():
            # to make Python > 3.5 behave the same as Python 3.5
            raise FileNotFoundError(self._source_folder)
        self._short_name = self._namespace_components_stropped[-1]
        self._data_type_to_outputs = dict()  # type: typing.Dict[pydsdl.CompositeType, pathlib.Path]
        self._nested_namespaces = set()  # type: typing.Set[Namespace]
        self._language_context = language_context

    @property
    def output_folder(self) -> pathlib.Path:
        """
        The folder where this namespace's output file and datatypes are generated.
        """
        return self._output_folder

    def get_support_output_folder(self) -> pathlib.PurePath:
        """
        The folder under which support artifacts are generated.
        """
        return self._support_output_folder

    def get_language_context(self) -> 'lang.LanguageContext':
        """
        The generated software language context the namespace is within.
        """
        return self._language_context

    def get_root_namespace(self) -> 'Namespace':
        """
        Traverses the namespace tree up to the root and returns the root node.

        :returns: The root namepace object.
        """
        namespace = self  # type: Namespace
        while namespace._parent is not None:
            namespace = namespace._parent
        return namespace

    def get_nested_namespaces(self) -> typing.Iterator['Namespace']:
        """
        Get an iterator over all the nested namespaces within this namespace.
        This is a shallow iterator that only provides directly nested namespaces.
        """
        return iter(self._nested_namespaces)

    def get_nested_types(self) -> typing.ItemsView[pydsdl.CompositeType, pathlib.Path]:
        """
        Get a view of a tuple relating datatypes in this namepace to the path for the
        type's generated output. This is a shallow view including only the types
        directly within this namespace.
        """
        return self._data_type_to_outputs.items()

    def get_all_datatypes(self) -> typing.Generator[typing.Tuple[pydsdl.CompositeType, pathlib.Path], None, None]:
        """
        Generates tuples relating datatypes at and below this namepace to the path
        for each type's generated output.
        """
        yield from self._recursive_data_type_generator(self)

    def get_all_namespaces(self) -> typing.Generator[typing.Tuple['Namespace', pathlib.Path], None, None]:
        """
        Generates tuples relating nested namespaces at and below this namepace to the path
        for each namespace's generated output.
        """
        yield from self._recursive_namespace_generator(self)

    def get_all_types(self) -> typing.Generator[typing.Tuple[pydsdl.Any, pathlib.Path], None, None]:
        """
        Generates tuples relating datatypes and nested namespaces at and below this
        namespace to the path for each type's generated output.
        """
        yield from self._recursive_data_type_and_namespace_generator(self)

    def find_output_path_for_type(self, any_type: pydsdl.Any) -> pathlib.Path:
        """
        Searches the entire namespace tree to find a mapping of the type to an
        output file path.

        :param Any any_type: Either a Namespace or pydsdl.CompositeType to find the
                             output path for.
        :returns: The path where a file will be generated for a given type.
        :raises KeyError: If the type was not found in this namespace tree.
        """
        if isinstance(any_type, Namespace):
            return any_type._output_path
        else:
            try:
                return self._data_type_to_outputs[any_type]
            except KeyError:
                pass

            # We could get fancier but this should do
            return self.get_root_namespace()._bfs_search_for_output_path(any_type, set([self]))

    # +-----------------------------------------------------------------------+
    # | DUCK TYPING: pydsdl.CompositeType
    # +-----------------------------------------------------------------------+
    @property
    def full_name(self) -> str:
        return self._full_namespace

    @property
    def full_namespace(self) -> str:
        return self._full_namespace

    @property
    def source_file_path(self) -> str:
        return str(self._source_folder)

    @property
    def data_types(self) -> typing.KeysView[pydsdl.CompositeType]:
        return self._data_type_to_outputs.keys()

    @property
    def attributes(self) -> typing.List[pydsdl.CompositeType]:
        return []

    # +-----------------------------------------------------------------------+
    # | PYTHON DATA MODEL
    # +-----------------------------------------------------------------------+

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Namespace):
            return self._full_namespace == other._full_namespace
        else:
            return False

    def __str__(self) -> str:
        return self.full_name

    def __hash__(self) -> int:
        return hash(self._full_namespace)

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+
    def _add_data_type(self, dsdl_type: pydsdl.CompositeType, extension: typing.Optional[str]) -> None:
        filestem = "{}_{}_{}".format(
            dsdl_type.short_name, dsdl_type.version.major, dsdl_type.version.minor)
        output_path = self._output_folder / pathlib.PurePath(filestem)
        if extension is not None:
            output_path = output_path.with_suffix(extension)
        self._data_type_to_outputs[dsdl_type] = output_path

    def _add_nested_namespace(self, nested: 'Namespace') -> None:
        self._nested_namespaces.add(nested)
        nested._parent = self

    def _bfs_search_for_output_path(self, data_type: pydsdl.CompositeType, skip_namespace: typing.Set['Namespace']) \
            -> pathlib.Path:
        search_queue = collections.deque()  # type: typing.Deque[Namespace]
        search_queue.appendleft(self)
        while len(search_queue) > 0:
            namespace = search_queue.pop()
            if namespace not in skip_namespace:
                try:
                    return namespace._data_type_to_outputs[data_type]
                except KeyError:
                    pass
            for nested_namespace in namespace._nested_namespaces:
                search_queue.appendleft(nested_namespace)

        raise KeyError(data_type)

    @classmethod
    def _recursive_data_type_generator(cls, namespace: 'Namespace') -> \
            typing.Generator[typing.Tuple[pydsdl.CompositeType, pathlib.Path], None, None]:
        for data_type, output_path in namespace.get_nested_types():
            yield (data_type, output_path)

        for nested_namespace in namespace.get_nested_namespaces():
            yield from cls._recursive_data_type_generator(nested_namespace)

    @classmethod
    def _recursive_namespace_generator(cls, namespace: 'Namespace') -> \
            typing.Generator[typing.Tuple['Namespace', pathlib.Path], None, None]:
        yield (namespace, namespace._output_path)

        for nested_namespace in namespace.get_nested_namespaces():
            yield from cls._recursive_namespace_generator(nested_namespace)

    @classmethod
    def _recursive_data_type_and_namespace_generator(cls, namespace: 'Namespace') -> \
            typing.Generator[typing.Tuple[pydsdl.Any, pathlib.Path], None, None]:
        yield (namespace, namespace._output_path)

        for data_type, output_path in namespace.get_nested_types():
            yield (data_type, output_path)

        for nested_namespace in namespace.get_nested_namespaces():
            yield from cls._recursive_data_type_and_namespace_generator(nested_namespace)

# +---------------------------------------------------------------------------+


def build_namespace_tree(types: typing.List[pydsdl.CompositeType],  # noqa: C901
                         root_namespace_dir: str,
                         output_dir: str,
                         language_context: lang.LanguageContext) -> Namespace:
    """Generates a :class:`nunavut.Namespace` tree.

    Given a list of pydsdl types, this method returns a root :class:`nunavut.Namespace`.
    The root :class:`nunavut.Namespace` is the top of a tree where each node contains
    references to nested :class:`nunavut.Namespace` and to any :code:`pydsdl.CompositeType`
    instances contained within the namespace.

    :param list types: A list of pydsdl types.
    :param str root_namespace_dir: A path to the folder which is the root namespace.
    :param str output_dir: The base directory under which all generated files will be created.
    :param lang.LanguageContext language_context: The language context to use when building
            :class:`nunavut.Namespace` objects.
    :returns: The root :class:`nunavut.Namespace`.

    """
    base_path = pathlib.PurePath(output_dir)

    namespace_index = set()  # type: typing.Set[str]
    namespaces = dict()  # type: typing.Dict[str, Namespace]

    def get_or_make_namespace(full_namespace: str) -> typing.Tuple[Namespace, bool]:
        # Local Namespace read through cache and factory.
        try:
            namespace = namespaces[str(full_namespace)]
            return (namespace, True)
        except KeyError:
            pass

        namespace = Namespace(full_namespace,
                              pathlib.Path(root_namespace_dir),
                              base_path,
                              language_context)

        namespaces[str(full_namespace)] = namespace

        return (namespace, False)

    for dsdl_type in types:
        # For each type we form a path with the output_dir as the base; the intermediate
        # folders named for the type's namespaces; and a file name that includes the type's
        # short name, major version, minor version, and the extension argument as a suffix.
        # Python's pathlib adapts the provided folder and file names to the platform
        # this script is running on.
        # We also, lazily, generate Namespace nodes as we encounter new namespaces for the
        # first time.

        namespace, did_exist = get_or_make_namespace(dsdl_type.full_namespace)

        if not did_exist:
            # add all namespaces up to root to index so we trigger
            # empty namespace generation in the final tree building
            # loop below.
            for i in range(len(dsdl_type.name_components) - 1, 0, -1):
                ancestor_ns = '.'.join(dsdl_type.name_components[0:i])
                # This little optimization pushed the complexity metric
                # too high which is why I did noqa here.
                if ancestor_ns in namespace_index:
                    break
                namespace_index.add(ancestor_ns)

        namespace._add_data_type(dsdl_type, language_context.get_output_extension())

    # We now have an index of all namespace names and we have Namespace
    # objects for non-empty namespaces. This final loop will build any
    # missing (i.e. empty) namespaces and all the links to form the
    # namespace tree.
    for full_namespace in namespace_index:
        namespace, _ = get_or_make_namespace(full_namespace)

        parent_namespace_components = namespace._namespace_components[0:-1]
        if (len(parent_namespace_components) > 0):
            parent_name = '.'.join(parent_namespace_components)

            parent, _ = get_or_make_namespace(parent_name)
            parent._add_nested_namespace(namespace)

    try:
        return next(iter(namespaces.values())).get_root_namespace()
    except StopIteration:
        pass

    # The empty namespace
    return get_or_make_namespace('')[0]

# +---------------------------------------------------------------------------+


class Dependencies:
    """
    Data structure that contains a set of composite types and annotations (bool flags)
    which constitute a set of dependencies for a set of DSDL objects.
    """

    def __init__(self) -> None:
        self.composite_types = set()  # type: typing.Set[pydsdl.CompositeType]
        self.uses_integer = False
        self.uses_float = False
        self.uses_variable_length_array = False
        self.uses_array = False
        self.uses_bool = False
        self.uses_primitive_static_array = False


class DependencyBuilder:
    """
    Given a list of DSDL types this object builds a set of types that the given types use.

    .. invisible-code-block: python

        import pydsdl
        from unittest.mock import MagicMock
        from nunavut import DependencyBuilder

        my_dependant_type_l2 = MagicMock(spec=pydsdl.CompositeType)
        my_dependant_type_l2.parent_service = False
        my_dependant_type_l2.attributes = []

        my_dependant_type_l1 = MagicMock(spec=pydsdl.CompositeType)
        my_dependant_type_l1.parent_service = False
        my_dependant_type_l1.attributes = [MagicMock(data_type=my_dependant_type_l2)]

        my_top_level_type = MagicMock(spec=pydsdl.UnionType)
        my_top_level_type.parent_service = False
        my_top_level_type.attributes = [MagicMock(data_type=my_dependant_type_l1)]

        direct_dependencies = DependencyBuilder(my_top_level_type).direct()

        assert len(direct_dependencies.composite_types) == 1
        assert my_dependant_type_l1 in direct_dependencies.composite_types

        transitive_dependencies = DependencyBuilder(my_top_level_type).transitive()

        assert len(transitive_dependencies.composite_types) == 2
        assert my_dependant_type_l1 in transitive_dependencies.composite_types
        assert my_dependant_type_l2 in transitive_dependencies.composite_types
        assert direct_dependencies.uses_integer

    :param dependant_types: A list of types to build dependencies for.
    :type dependant_types: typing.Iterable[pydsdl.Any]
    """

    def __init__(self, *dependant_types: pydsdl.Any):
        self._dependent_types = dependant_types

    def transitive(self) -> Dependencies:
        """
        Build a set of all transitive dependencies for the dependent types
        set for this builder.
        """
        return self._build_dependency_list(self._dependent_types, True)

    def direct(self) -> Dependencies:
        """
        Build a set of all first-order dependencies for the dependent types
        set for this builder.
        """
        return self._build_dependency_list(self._dependent_types, False)

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+

    @classmethod
    def _build_dependency_list(cls, dependant_types: typing.Iterable[pydsdl.CompositeType], transitive: bool) \
            -> Dependencies:
        results = Dependencies()
        for dependant in dependant_types:
            if isinstance(dependant, pydsdl.UnionType):
                # Unions always require integer for the tag field.
                results.uses_integer = True
            cls._extract_dependent_types(cls._extract_data_types(dependant), transitive, results)
        return results

    @classmethod
    def _extract_data_types(cls, t: pydsdl.CompositeType) -> typing.List[pydsdl.SerializableType]:
        # Make a list of all attributes defined by this type
        if isinstance(t, pydsdl.ServiceType):
            return [attr.data_type for attr in t.request_type.attributes] + \
                [attr.data_type for attr in t.response_type.attributes]
        else:
            return [attr.data_type for attr in t.attributes]

    @classmethod
    def _extract_dependent_types_handle_array_type(cls,
                                                   dependant_type: pydsdl.ArrayType,
                                                   transitive: bool,
                                                   inout_dependencies: Dependencies) -> None:
        if isinstance(dependant_type, pydsdl.VariableLengthArrayType):
            inout_dependencies.uses_variable_length_array = True
        else:
            inout_dependencies.uses_array = True
            if isinstance(dependant_type.element_type, pydsdl.PrimitiveType):
                inout_dependencies.uses_primitive_static_array = True

    @classmethod
    def _extract_dependent_types(cls,
                                 dependant_types: typing.Iterable[pydsdl.Any],
                                 transitive: bool,
                                 inout_dependencies: Dependencies) -> None:
        for dt in dependant_types:
            if isinstance(dt, pydsdl.CompositeType):
                if dt not in inout_dependencies.composite_types:
                    inout_dependencies.composite_types.add(dt)
                    if transitive:
                        cls._extract_dependent_types(cls._extract_data_types(dt), transitive, inout_dependencies)
            elif isinstance(dt, pydsdl.ArrayType):
                cls._extract_dependent_types_handle_array_type(dt, transitive, inout_dependencies)
                cls._extract_dependent_types([dt.element_type], transitive, inout_dependencies)
            elif isinstance(dt, pydsdl.IntegerType):
                inout_dependencies.uses_integer = True
            elif isinstance(dt, pydsdl.FloatType):
                inout_dependencies.uses_float = True
            elif isinstance(dt, pydsdl.BooleanType):
                inout_dependencies.uses_bool = True

# +---------------------------------------------------------------------------+
# | GENERATION HELPERS
# +---------------------------------------------------------------------------+


def generate_types(language_key: str,
                   root_namespace_dir: pathlib.Path,
                   out_dir: pathlib.Path,
                   omit_serialization_support: bool = True,
                   is_dryrun: bool = False,
                   allow_overwrite: bool = True,
                   lookup_directories: typing.Optional[typing.Iterable[str]] = None,
                   allow_unregulated_fixed_port_id: bool = False,
                   language_options: typing.Optional[typing.Mapping[str, typing.Any]] = None) -> None:
    """
    Helper method that uses default settings and built-in templates to generate types for a given
    language. This method is the most direct way to generate code using Nunavut.

    :param str language_key: The name of the language to generate source for.
                See the :doc:`../../docs/templates` for details on available language support.
    :param pathlib.Path root_namespace_dir: The path to the root of the DSDL types to generate
                code for.
    :param pathlib.Path out_dir: The path to generate code at and under.
    :param bool omit_serialization_support: If True then logic used to serialize and deserialize data is omitted.
    :param bool is_dryrun: If True then nothing is generated but all other activity is performed and any errors
                that would have occurred are reported.
    :param bool allow_overwrite: If True then generated files are allowed to overwrite existing files under the
                `out_dir` path.
    :param typing.Optional[typing.Iterable[str]] lookup_directories: Additional directories to search for dependent
                types referenced by the types provided under the `root_namespace_dir`. Types will not be generated
                for these unless they are used by a type in the root namespace.
    :param bool allow_unregulated_fixed_port_id: If True then errors will become warning when using fixed port
                identifiers for unregulated datatypes.
    :param typing.Optional[typing.Mapping[str, typing.Any]] language_options: Opaque arguments passed through to the
                language objects. The supported arguments and valid values are different depending on the language
                specified by the `language_key` parameter.
    """
    from nunavut.generators import create_generators

    language_context = lang.LanguageContext(language_key,
                                            omit_serialization_support_for_target=omit_serialization_support,
                                            language_options=language_options)

    if lookup_directories is None:
        lookup_directories = []

    type_map = pydsdl.read_namespace(str(root_namespace_dir),
                                     lookup_directories,
                                     allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id)

    namespace = build_namespace_tree(type_map,
                                     str(root_namespace_dir),
                                     str(out_dir),
                                     language_context)

    generator, support_generator = create_generators(namespace)
    support_generator.generate_all(is_dryrun, allow_overwrite)
    generator.generate_all(is_dryrun, allow_overwrite)



from functools import wraps

from nunavut.jinja.jinja2.asyncsupport import auto_aiter, auto_await
from nunavut.jinja.jinja2 import filters


async def auto_to_seq(value):
    seq = []
    if hasattr(value, '__aiter__'):
        async for item in value:
            seq.append(item)
    else:
        for item in value:
            seq.append(item)
    return seq


async def async_select_or_reject(args, kwargs, modfunc, lookup_attr):
    seq, func = filters.prepare_select_or_reject(
        args, kwargs, modfunc, lookup_attr)
    if seq:
        async for item in auto_aiter(seq):
            if func(item):
                yield item


def dualfilter(normal_filter, async_filter):
    wrap_evalctx = False
    if getattr(normal_filter, 'environmentfilter', False):
        is_async = lambda args: args[0].is_async
        wrap_evalctx = False
    else:
        if not getattr(normal_filter, 'evalcontextfilter', False) and \
           not getattr(normal_filter, 'contextfilter', False):
            wrap_evalctx = True
        is_async = lambda args: args[0].environment.is_async

    @wraps(normal_filter)
    def wrapper(*args, **kwargs):
        b = is_async(args)
        if wrap_evalctx:
            args = args[1:]
        if b:
            return async_filter(*args, **kwargs)
        return normal_filter(*args, **kwargs)

    if wrap_evalctx:
        wrapper.evalcontextfilter = True

    wrapper.asyncfiltervariant = True

    return wrapper


def asyncfiltervariant(original):
    def decorator(f):
        return dualfilter(original, f)
    return decorator


@asyncfiltervariant(filters.do_first)
async def do_first(environment, seq):
    try:
        return await auto_aiter(seq).__anext__()
    except StopAsyncIteration:
        return environment.undefined('No first item, sequence was empty.')


@asyncfiltervariant(filters.do_groupby)
async def do_groupby(environment, value, attribute):
    expr = filters.make_attrgetter(environment, attribute)
    return [filters._GroupTuple(key, await auto_to_seq(values))
            for key, values in filters.groupby(sorted(
                await auto_to_seq(value), key=expr), expr)]


@asyncfiltervariant(filters.do_join)
async def do_join(eval_ctx, value, d=u'', attribute=None):
    return filters.do_join(eval_ctx, await auto_to_seq(value), d, attribute)


@asyncfiltervariant(filters.do_list)
async def do_list(value):
    return await auto_to_seq(value)


@asyncfiltervariant(filters.do_reject)
async def do_reject(*args, **kwargs):
    return async_select_or_reject(args, kwargs, lambda x: not x, False)


@asyncfiltervariant(filters.do_rejectattr)
async def do_rejectattr(*args, **kwargs):
    return async_select_or_reject(args, kwargs, lambda x: not x, True)


@asyncfiltervariant(filters.do_select)
async def do_select(*args, **kwargs):
    return async_select_or_reject(args, kwargs, lambda x: x, False)


@asyncfiltervariant(filters.do_selectattr)
async def do_selectattr(*args, **kwargs):
    return async_select_or_reject(args, kwargs, lambda x: x, True)


@asyncfiltervariant(filters.do_map)
async def do_map(*args, **kwargs):
    seq, func = filters.prepare_map(args, kwargs)
    if seq:
        async for item in auto_aiter(seq):
            yield await auto_await(func(item))


@asyncfiltervariant(filters.do_sum)
async def do_sum(environment, iterable, attribute=None, start=0):
    rv = start
    if attribute is not None:
        func = filters.make_attrgetter(environment, attribute)
    else:
        func = lambda x: x
    async for item in auto_aiter(iterable):
        rv += func(item)
    return rv


@asyncfiltervariant(filters.do_slice)
async def do_slice(value, slices, fill_with=None):
    return filters.do_slice(await auto_to_seq(value), slices, fill_with)


ASYNC_FILTERS = {
    'first':        do_first,
    'groupby':      do_groupby,
    'join':         do_join,
    'list':         do_list,
    # we intentionally do not support do_last because that would be
    # ridiculous
    'reject':       do_reject,
    'rejectattr':   do_rejectattr,
    'map':          do_map,
    'select':       do_select,
    'selectattr':   do_selectattr,
    'sum':          do_sum,
    'slice':        do_slice,
}



#
# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""Language-specific support in nunavut.

This package contains modules that provide specific support for generating
source for various languages using templates.
"""
import functools
import importlib
import inspect
import logging
import pathlib
import typing

from ._config import LanguageConfig

logger = logging.getLogger(__name__)


class Language:
    """
    Facilities for generating source code for a specific language.

    :param str language_name:                The name of the language used by the :mod:`nunavut.lang` module.
    :param LanguageConfig config:            The parser to load language properties into.
    :param bool omit_serialization_support:  The value to set for the :func:`omit_serialization_support` property
                                             for this language.
    :param typing.Optional[typing.Mapping[str, typing.Any]] language_options: Opaque arguments passed through to the
                target :class:`nunavut.lang.Language` object.
    """

    def _find_callable_for_language(self, callable_name_prefix: str) -> \
            typing.Mapping[str, typing.Callable]:
        callable_map = dict()  # type: typing.Dict[str, typing.Callable]
        lang_module = self._language_module
        callables = inspect.getmembers(lang_module, inspect.isfunction)
        callable_name_prefix_len = len(callable_name_prefix)
        for function_tuple in callables:
            function_name = function_tuple[0]
            if len(function_name) > callable_name_prefix_len and \
                    function_name[0:callable_name_prefix_len] == callable_name_prefix:
                callable_map[function_name[callable_name_prefix_len:]] = function_tuple[1]
                logging.debug("Found callable {} for language {}".format(function_name[callable_name_prefix_len:],
                                                                         self._language_name))
        return callable_map

    def _find_filters_for_language(self) -> typing.Mapping[str, typing.Callable]:
        return self._find_callable_for_language('filter_')

    def _find_tests_for_language(self) -> typing.Mapping[str, typing.Callable]:
        return self._find_callable_for_language('is_')

    @classmethod
    def get_language_config_parser(cls) -> LanguageConfig:
        """
        Create a :class:`LanguageConfig` instance configured for reading
        language properties.yaml.
        """
        return LanguageConfig()

    def __init__(self,
                 language_name: str,
                 config: LanguageConfig,
                 omit_serialization_support: bool,
                 language_options: typing.Optional[typing.Mapping[str, typing.Any]] = None):
        self._globals = None  # type: typing.Optional[typing.Mapping[str, typing.Any]]
        self._language_name = language_name
        self._section = 'nunavut.lang.{}'.format(language_name)
        self._language_module = importlib.import_module(self._section)
        self._config_getter_cache = {}  # type: typing.Mapping[str, functools.partial[typing.Any]]
        self._filters = self._find_filters_for_language()
        self._tests = self._find_tests_for_language()
        self._config = config
        self._omit_serialization_support = omit_serialization_support
        self._language_options = config.get_config_value_as_dict(self._section, 'options', dict())

        if language_options is not None:
            self._language_options.update(language_options)

    def __getattr__(self, name: str) -> typing.Any:
        """
        Any attribute access to a Language object will return the regular properties and
        any globals defined for the language. Because of this do not extend properties
        on this object in a way that will clash with the globals it defines (e.g. typename_
        or valuetoken_ should not be used as the start of attribute names).
        """
        try:
            return self.get_globals()[name]
        except KeyError as e:
            raise AttributeError(e)

    def get_config_value(self,
                         key: str,
                         default_value: typing.Optional[str] = None) -> str:
        """
        Get an optional language property from the language configuration.

        :param str key: The config value to retrieve.
        :param default_value: The value to return if the key was not in the configuration. If provided
                this method will not raise.
        :type default_value: typing.Optional[str]
        :return: Either the value from the config or the default_value if provided.
        :rtype: str
        :raises: KeyError if the section or the key in the section does not exist and a default_value was not provided.
        """
        return self._config.get_config_value(self._section, key, default_value)

    def get_config_value_as_bool(self, key: str, default_value: bool = False) -> bool:
        """
        Get an optional language property from the language configuration returning a boolean.

        :param str key: The config value to retrieve.
        :param bool default_value: The value to use if no value existed.
        :return: The config value as either True or False.
        :rtype: bool
        """
        return self._config.get_config_value_as_bool(self._section, key, default_value)

    def get_config_value_as_dict(self, key: str, default_value: typing.Optional[typing.Dict] = None) -> \
            typing.Dict[str, typing.Any]:
        """
        Get a language property parsing it as a map with string keys.

        .. invisible-code-block: python

            from nunavut.lang import LanguageConfig, Language

            config = LanguageConfig()
            config.add_section('nunavut.lang.c')
            config.set('nunavut.lang.c', 'foo', {'one': 1})

            lang_c = Language('c', config, True)

            assert lang_c.get_config_value_as_dict('foo')['one'] == 1

            assert lang_c.get_config_value_as_dict('bar', {'one': 2})['one'] == 2

        :param str key:           The config value to retrieve.
        :param default_value:     The value to return if the key was not in the configuration. If provided this method
            will not raise a KeyError nor a TypeError.
        :type default_value: typing.Optional[typing.Mapping[str, typing.Any]]
        :return:                  Either the value from the config or the default_value if provided.
        :rtype: typing.Mapping[str, typing.Any]
        :raises: KeyError if the key does not exist and a default_value was not provided.
        :raises: TypeError if the value exists but is not a dict and a default_value was not provided.

        """
        return self._config.get_config_value_as_dict(self._section, key, default_value)

    def get_config_value_as_list(self, key: str, default_value: typing.Optional[typing.List] = None)\
            -> typing.List[typing.Any]:
        """
        Get a language property parsing it as a map with string keys.

        :param str key:           The config value to retrieve.
        :param default_value:     The value to return if the key was not in the configuration. If provided this method
            will not raise a KeyError nor a TypeError.
        :type default_value:      typing.Optional[typing.List[typing.Any]]
        :return:                  Either the value from the config or the default_value if provided.
        :rtype:                   typing.List[typing.Any]
        :raises:                  KeyError if the key does not exist and a default_value was not provided.
        :raises:                  TypeError if the value exists but is not a dict and a default_value was not provided.

        """
        return self._config.get_config_value_as_list(self._section, key, default_value)

    @property
    def extension(self) -> str:
        """
        The extension to use for files generated in this language.
        """
        return self._config.get_config_value(self._section, 'extension', 'get')

    @property
    def namespace_output_stem(self) -> typing.Optional[str]:
        """
        The name of a namespace file for this language.
        """
        try:
            return self._config.get_config_value(self._section, 'namespace_file_stem')
        except KeyError:
            return None

    @property
    def name(self) -> str:
        """
        The name of the language used by the :mod:`nunavut.lang` module.
        """
        return self._language_name

    @property
    def support_namespace(self) -> typing.List[str]:
        """
        The hierarchical namespace used by the support software. The property
        is a dot separated string when specified in configuration. This
        property returns that value split into namespace components with the
        first identifier being the first index in the array, etc.

        .. invisible-code-block: python

            from nunavut.lang import Language

            config = Language.get_language_config_parser()
            config.add_section('nunavut.lang.cpp')

            lang_cpp = Language('cpp', config, True)

        .. code-block:: python

            config.set('nunavut.lang.cpp', 'support_namespace', 'foo.bar')
            assert len(lang_cpp.support_namespace) == 2
            assert lang_cpp.support_namespace[0] == 'foo'
            assert lang_cpp.support_namespace[1] == 'bar'

        """
        namespace_str = self._config.get_config_value(self._section, 'support_namespace', default_value='')
        return namespace_str.split('.')

    @property
    def enable_stropping(self) -> bool:
        """
        Whether or not to strop identifiers for this language.
        """
        return self._config.get_config_value_as_bool(self._section, 'enable_stropping')

    @property
    def has_standard_namespace_files(self) -> bool:
        """
        Whether or not the language defines special namespace files as part of
        its core standard (e.g. python's __init__).
        """
        return self._config.get_config_value_as_bool(self._section, 'has_standard_namespace_files')

    @property
    def stable_support(self) -> bool:
        """
        Whether support for this language is designated 'stable', and not experimental.
        """
        return self._config.get_config_value_as_bool(self._section, 'stable_support')

    @property
    def omit_serialization_support(self) -> bool:
        """
        If True then generators should not include serialization routines, types,
        or support libraries for this language.
        """
        return self._omit_serialization_support

    @property
    def support_files(self) -> typing.Generator[pathlib.Path, None, None]:
        """
        Iterates over non-templated supporting files embedded within the Nunavut distribution.

        .. invisible-code-block: python

            from nunavut.lang import Language
            from unittest.mock import MagicMock

            mock_config = MagicMock()

            my_lang = Language('c', mock_config, True)
            my_lang._section = "nunavut.lang.not_a_language_really_not_a_language"
            for support_file in my_lang.support_files:
                # if the module doesn't exist it shouldn't have any support files.
                assert False

        """
        try:
            module = importlib.import_module('{}.support'.format(self._section))

            # All language support modules must provide a list_support_files method
            # to allow the copy generator access to the packaged support files.
            list_support_files = getattr(module, 'list_support_files')  \
                # type: typing.Callable[[], typing.Generator[pathlib.Path, None, None]]
            return list_support_files()
        except ImportError:
            # No serialization support for this language
            logger.info("No serialization support for selected target. Skipping serialization support generation.")

            def list_support_files() -> typing.Generator[pathlib.Path, None, None]:
                # This makes both MyPy and sonarqube happy.
                return typing.cast(typing.Generator[pathlib.Path, None, None], iter(()))

            return list_support_files()

    def get_option(self,
                   option_key: str,
                   default_value: typing.Union[typing.Mapping[str, typing.Any], str, None] = None)\
            -> typing.Union[typing.Mapping[str, typing.Any], str, None]:
        """
        Get a language option for this language.

        .. invisible-code-block: python

            from nunavut.lang import Language

            config = Language.get_language_config_parser()
            config.add_section('nunavut.lang.cpp')
            config.set('nunavut.lang.cpp', 'options', {'target_endianness': 'little'})

            lang_cpp = Language('cpp', config, True)

        .. code-block:: python

            # Values can come from defaults...
            assert lang_cpp.get_option('target_endianness') == 'little'

            # ... or can be provided to a language instance.
            my_lang = Language('cpp', config, True, language_options={'target_endianness': 'any'})
            assert my_lang.get_option('target_endianness') == 'any'

            # Also, this method can provide a sane default.
            assert lang_cpp.get_option('foobar', 'sane_default') == 'sane_default'

        :return: Either the value provided to the :class:`Language` instance, the value from properties.yaml,
            or the :code:`default_value`.

        """
        try:
            return self._language_options[option_key]  # type: ignore
        except KeyError:
            return default_value

    def get_templates_package_name(self) -> str:
        """
        The name of the nunavut python package containing filters, types, and configuration
        for this language.
        """
        return self._section

    def get_named_types(self) -> typing.Mapping[str, str]:
        """
        Get a map of named types to the type name to emit for this language.
        """
        return self._config.get_config_value_as_dict(self._section, 'named_types', default_value={})

    def get_named_values(self) -> typing.Mapping[str, str]:
        """
        Get a map of named values to the token to emit for this language.
        """
        return self._config.get_config_value_as_dict(self._section, 'named_values', default_value={})

    def get_filters(self) -> typing.Mapping[str, typing.Callable]:
        """
        Inspect the language module for functions with a name starting with "filter\\_" and return
        a map of filter names to the filter callable.

        :returns: A mapping of filter names to filter functions.
        """
        return self._filters

    def get_tests(self) -> typing.Mapping[str, typing.Callable]:
        """
        Inspect the language module for functions with a name starting with "is\\_" and return
        a map of test names to the test callable.

        :returns: A mapping of test names to test functions.
        """
        return self._tests

    def get_globals(self) -> typing.Mapping[str, typing.Any]:
        """
        Get all values for this language that should be available in a global context.

        :returns: A mapping of global names to global values.
        """
        if self._globals is None:
            globals_map = dict()  # type: typing.Dict[str, typing.Any]

            for key, value in self.get_named_types().items():
                globals_map['typename_{}'.format(key)] = value
            for key, value in self.get_named_values().items():
                globals_map['valuetoken_{}'.format(key)] = value

            self._globals = globals_map
        return self._globals

    def get_options(self) -> typing.Mapping[str, typing.Any]:
        """
        Get all language options for this Language.

        :returns: A mapping of option names to option values.
        """
        return self._language_options


class LanguageContext:
    """
    Context object containing the current target language (if any) and used to access
    :class:`Language` objects.

    :param str target_language: If provided a :class:`Language` object will be created to hold
                                the target language set for this context. If None then there is
                                no target language.
    :param str extension: The extension to use for generated file types or None to use a default
                          based on the target_language.
    :param str namespace_output_stem: The filename stem to give to Namespace output files if
                                      emitted or None to use a default based on a target_language.
    :param additional_config_files: A list of paths to additional ini files to load as configuration.
        These will override any values found in the :file:`nunavut.lang.properties.yaml` file and files
        appearing later in this list will override value found in earlier entries.
    :type additional_config_files: typing.List[pathlib.Path]
    :param bool omit_serialization_support_for_target: If True then generators should not include
        serialization routines, types, or support libraries for the target language.
    :param typing.Optional[typing.Mapping[str, typing.Any]] language_options: Opaque arguments passed through to the
                target :class:`nunavut.lang.Language` object.
    :param bool include_experimental_languages: If True, expose languages with experimental (non-stable) support.
    :raises ValueError: If extension is None and no target language was provided.
    :raises KeyError: If the target language is not known.
    """

    @classmethod
    def _load_config(cls, *additional_config_files: pathlib.Path) -> LanguageConfig:
        import pkg_resources
        parser = Language.get_language_config_parser()
        resources = [r for r in pkg_resources.resource_listdir(__name__, '.') if r.endswith('.yaml')]
        for resource in resources:
            with pkg_resources.resource_stream(__name__, resource) as resource_stream:
                ini_string = resource_stream.read().decode(encoding='utf-8', errors='replace')
                parser.read_string(ini_string, resource)
        for additional_path in additional_config_files:
            with open(str(additional_path), 'r') as additional_file:
                parser.read_file(additional_file)
        return parser

    def __init__(self,
                 target_language: typing.Optional[str] = None,
                 extension: typing.Optional[str] = None,
                 namespace_output_stem: typing.Optional[str] = None,
                 additional_config_files: typing.List[pathlib.Path] = [],
                 omit_serialization_support_for_target: bool = True,
                 language_options: typing.Optional[typing.Mapping[str, typing.Any]] = None,
                 include_experimental_languages: bool = True):
        self._extension = extension
        self._namespace_output_stem = namespace_output_stem
        self._config = self._load_config(*additional_config_files)
        self._languages = dict()  # type: typing.Dict[str, Language]

        # create target language, if there is one.
        self._target_language = None
        if target_language is not None:
            try:
                self._target_language = Language(target_language, self._config,
                                                 omit_serialization_support_for_target,
                                                 language_options=language_options)
            except ImportError:
                raise KeyError('{} is not a supported language'.format(target_language))
            if not (self._target_language.stable_support or include_experimental_languages):
                raise ValueError(
                    '{} support is only experimental, but experimental language support is not enabled'
                    .format(target_language)
                )
            if namespace_output_stem is None:
                self._namespace_output_stem = self._target_language.namespace_output_stem

            target_language_section_name = 'nunavut.lang.{}'.format(target_language)
            if self._namespace_output_stem is not None:
                self._config.set(target_language_section_name,
                                 'namespace_file_stem',
                                 self._namespace_output_stem)
            if extension is not None:
                self._config.set(target_language_section_name,
                                 'extension',
                                 extension)
            self._languages[target_language] = self._target_language

        # create remaining languages
        remaining_languages = set(self.get_supported_language_names()) - set((target_language,))
        self._populate_languages(remaining_languages, include_experimental_languages)

    def _populate_languages(self, language_names: typing.Iterable[str],
                            include_experimental: bool) -> None:
        for language_name in language_names:
            try:
                lang = Language(language_name, self._config, False)
                if lang.stable_support or include_experimental:
                    self._languages[language_name] = lang
            except ImportError:
                raise KeyError('{} is not a supported language'.format(language_name))

    def get_language(self, key_or_module_name: str) -> Language:
        """
        Get a :class:`Language` object for a given language identifier.

        :param str key_or_module_name: Either one of the Nunavut mnemonics for a supported language or
            the ``__name__`` of one of the ``nunavut.lang.[language]`` python modules.
        :returns: A :class:`Language` object cached by this context.
        :rtype: Language
        """
        if key_or_module_name is None or len(key_or_module_name) == 0:
            raise ValueError('key argument is required.')
        key = (key_or_module_name[13:] if key_or_module_name.startswith('nunavut.lang.') else key_or_module_name)
        return self.get_supported_languages()[key]

    def get_supported_language_names(self) -> typing.Iterable[str]:
        """Get a list of target languages supported by Nunavut.

        :returns: An iterable of strings which are languages with special
            support within Nunavut templates.
        """
        return [s[13:] for s in self._config.sections() if s.startswith('nunavut.lang.')]

    def get_output_extension(self) -> str:
        """
        Gets the output extension to use regardless of a target language being available or not.

        :returns: A file extension name with a leading dot.
        """
        if self._extension is not None:
            return self._extension
        elif self._target_language is not None:
            return self._target_language.extension
        else:
            raise RuntimeError('No extension was provided and no target language was set.'
                               'Cannot determine the extension to use.')

    def get_default_namespace_output_stem(self) -> typing.Optional[str]:
        """
        The filename stem to give to Namespace output files if emitted or None if there was none
        specified and there is no target language.

        :returns: A file name stem or None
        """
        return self._namespace_output_stem

    def get_target_language(self) -> typing.Optional[Language]:
        """
        Returns the target language configured on this object or None
        if no target language was specified.
        """
        return self._target_language

    def get_target_id_filter(self) -> typing.Callable[[str, str], str]:
        """
        A filter that will transform a given string into a valid identifier
        in the target language. The string is pass through unmodified if no
        target language was set.
        """
        if self._target_language is not None:
            filters = self._target_language.get_filters()
            for name, filter in filters.items():
                if name == 'id':
                    id_filter = typing.cast(typing.Callable[[Language, str, str], str], filter)
                    return functools.partial(id_filter, self._target_language)

        def no_filter(unfiltered: str, _: str = '') -> str:
            return unfiltered

        return no_filter

    def get_supported_languages(self) -> typing.Dict[str, Language]:
        """
        Returns a collection of available language support objects.
        """
        return self._languages

    @property
    def config(self) -> LanguageConfig:
        return self._config



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Contains the Jinja templates to generate C++ headers.
"""

__version__ = "1.0.0"
"""Version of the c++ template set."""



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Abstractions around template engine internals.
"""
import typing

ENVIRONMENT_FILTER_ATTRIBUTE_NAME = 'environmentfilter'
"""
For now this is set to a value that is internally compatible with the
embedded version of jinja2 we use in Nunavut. If you use this variable
instead of its current value you will be insulated from this.
"""

CONTEXT_FILTER_ATTRIBUTE_NAME = 'contextfilter'
"""
For now this is set to a value that is internally compatible with the
embedded version of jinja2 we use in Nunavut. If you use this variable
instead of its current value you will be insulated from this.
"""

LANGUAGE_FILTER_ATTRIBUTE_NAME = 'nv_languagefilter'
"""
Nunavut-specific attribute for filters or tests that take their :class:`nunavut.lang.Language`
as the first argument.
"""


class SupportsTemplateEnv:
    """
    Provided as a pseudo
    `protocol <https://mypy.readthedocs.io/en/latest/protocols.html#simple-user-defined-protocols/>`_.
    (in anticipation of that becoming part of the core Python typing someday).

    :var Dict globals: A dictionary mapping global names (str) to variables (Any) that are available in
                       each template's global environment.
    :var Dict filters: A dictionary mapping filter names (str) to filters (Callable). Filters are simply
                       global functions available to templates but are most often used to transform a
                       given input to output emitted by the template.
    """

    globals = dict()  # type: typing.Dict[str, typing.Any]
    filters = dict()  # type: typing.Dict[str, typing.Callable]
    tests = dict()  # type: typing.Dict[str, typing.Callable[[typing.Any], bool]]


class SupportsTemplateContext:
    """
    Provided as a pseudo
    `protocol <https://mypy.readthedocs.io/en/latest/protocols.html#simple-user-defined-protocols/>`_.
    (in anticipation of that becoming part of the core Python typing someday).
    """


def template_environment_filter(filter_func: typing.Callable) -> typing.Callable[..., str]:
    """
    Decorator for marking environment dependent filters.
    An object supporting the :class:`SupportsTemplateEnv` protocol
    will be passed to the filter as the first argument.
    """
    setattr(filter_func, ENVIRONMENT_FILTER_ATTRIBUTE_NAME, True)
    return filter_func


def template_context_filter(filter_func: typing.Callable) -> typing.Callable[..., str]:
    """
    Decorator for marking context dependent filters.
    An object supporting the :class:`SupportsTemplateContext` protocol
    will be passed to the filter as the first argument.
    """
    setattr(filter_func, CONTEXT_FILTER_ATTRIBUTE_NAME, True)
    return filter_func


LanguageFilterReturnType = typing.TypeVar('LanguageFilterReturnType')


class GenericTemplateLanguageFilter(typing.Generic[LanguageFilterReturnType]):
    """
    Decorator for marking template filters that take a :class:`nunavut.lang.Language` object
    as the first argument with a generic return type.
    """

    def __init__(self, language_name_or_module: str):
        self._language_name_or_module = language_name_or_module

    def __call__(self, filter_func: typing.Callable[..., LanguageFilterReturnType]) \
            -> typing.Callable[..., LanguageFilterReturnType]:
        self._annotate_function(filter_func)
        return filter_func

    def _annotate_function(self, filter_func: typing.Callable[..., typing.Any]) -> None:
        setattr(filter_func, LANGUAGE_FILTER_ATTRIBUTE_NAME, self._language_name_or_module)


class template_language_filter(GenericTemplateLanguageFilter[str]):
    """
    Decorator for marking template filters that take a :class:`nunavut.lang.Language` object
    as the first argument.
    """
    pass


class template_language_list_filter(GenericTemplateLanguageFilter[typing.List[str]]):
    """
    Decorator for marking template filters that take a :class:`nunavut.lang.Language` object
    as the first argument and return a list of strings.
    """
    pass


class template_language_int_filter(GenericTemplateLanguageFilter[int]):
    """
    Decorator for marking template filters that take a :class:`nunavut.lang.Language` object
    as the first argument and return an integer.
    """
    pass


class template_language_test(GenericTemplateLanguageFilter[bool]):
    """
    Decorator for marking template tests that take a :class:`nunavut.lang.Language` object
    as the first argument.
    """
    pass
